/*********************************************************************/
/*                          Litmus                                   */
/*                                                                   */
/*        Luc Maranget, INRIA Paris-Rocquencourt, France.            */
/*        Susmit Sarkar, University of Cambridge, UK.                */
/*                                                                   */
/*  Copyright 2010 Institut National de Recherche en Informatique et */
/*  en Automatique and the authors. All rights reserved.             */
/*  This file is distributed  under the terms of the Lesser GNU      */
/*  General Public License.                                          */
/*********************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <pthread.h>
#include <limits.h>
#include "utils.h"

/********/
/* Misc */
/********/

void err(int e, char *msg) {
  perror(msg) ;
  fprintf(stdout,"Failure: %s\n", msg) ;
  exit(e) ;
}


void *malloc_check(size_t sz) {
  if (sz == 0) return NULL ;
  void *p = malloc(sz) ;
  if (!p) err(1,"malloc") ;
  return p ;
}

/************/
/* CPU sets */
/************/

cpus_t *cpus_create(int sz) {
  cpus_t *r = malloc_check(sizeof(*r)) ;
  r->sz = sz ;
  r->cpu = malloc_check(sizeof(r->cpu[0])*sz)  ;
  return r ;
}

cpus_t *cpus_create_init(int sz, int t[]) {
  cpus_t *r = cpus_create(sz) ;
  for (int k = 0 ; k < sz ; k++) r->cpu[k] = t[k] ;
  return r ;
}

void cpus_free(cpus_t *p) {
  free(p->cpu) ;
  free(p) ;
}

void cpus_dump(FILE *fp, cpus_t *p) {
  if (p->sz > 0) {
    fprintf(fp,"%i",p->cpu[0]) ;
    for (int k = 1 ; k < p->sz ; k++) {
      fprintf(fp,",%i",p->cpu[k]) ;
    }
  }
}

int gcd(int a, int b) {
  for ( ; ; ) {
    if (a == 0) return b ;
    int tmp = a ;
    a = b % a ;
    b = tmp ;
  }
}


/****************/
/* Command line */
/****************/

/* usage */

static void usage(char *prog, cmd_t *d) {
  fprintf(stderr,"usage: %s (options)*\n",prog) ;
  fprintf(stderr,"  -v      be verbose\n") ;
  fprintf(stderr,"  -q      be quiet\n") ;
  fprintf(stderr,"  -a <n>  run maximal number of tests for n available processors (default %i)\n",d->avail) ;
  fprintf(stderr,"  -n <n>  run n tests concurrently\n") ;
  fprintf(stderr,"  -r <n>  perform n runs (default %i)\n",d->max_run) ;
  fprintf(stderr,"  -fr <f> multiply run number per f\n") ;
  fprintf(stderr,"  -s <n>  outcomes per run (default %i)\n",d->size_of_test) ;
  fprintf(stderr,"  -fs <f> multiply outcomes per f\n") ;
  fprintf(stderr,"  -f <f>  multiply outcomes per f, divide run number by f\n") ;
  if (d->aff_incr >= 0) {
    fprintf(stderr,"  -i <n>  increment for allocating logical processors, -i 0 disables affinity mode (default %i)\n",d->aff_incr) ;
    fprintf(stderr,"  -p <ns> specify logical processors (default '") ;
    cpus_dump(stderr,d->aff_cpus) ;
    fprintf(stderr,"')\n") ;
  }
  
  fprintf(stderr,"  -k <n>  undocumented (default %i)\n",d->sync_n) ;
  exit(2) ;
}

static int argint(char *prog,char *p,cmd_t *d) {
  char *q ;
  int r = strtol(p,&q,10) ;
  if (*p == '\0' || *q != '\0') {
    usage(prog,d) ;
  }
  return r ;
}

static cpus_t *argcpus(char *prog,char *p0,cmd_t *d) {
  int sz = 0 ;
  char *p ;

  p = p0 ;
  for ( ; ; ) {
    char *q ;
    int x = strtol(p,&q,10) ;
    if (x < 0 || *p == '\0' || (*q != '\0' && *q != ','))  usage(prog,d) ;
    sz++ ;
    if (*q == '\0') break ;
    p = q+1 ;    
  }
  cpus_t *r = cpus_create(sz) ;
  p = p0 ;
  for (int k = 0 ; k < sz ; k++) {
    char *q ;
    r->cpu[k] = strtol(p,&q,10) ;
    p = q+1 ;
  }
  return r ;
}


static double argdouble(char *prog,char *p,cmd_t *d) {
  char *q ;
  double r = strtod(p,&q) ;
  if (*p == '\0' || *q != '\0') {
    usage(prog,d) ;
  }
  return r ;
}

void parse_cmd(int argc, char **argv, cmd_t *d, cmd_t *p) {
  char *prog = argv[0] ;

  /* Options */
  for ( ; ; ) {
    --argc ; ++argv ;
    if (!*argv) break ;
    char fst = **argv ;
    if (fst != '-') break ;
    if (strcmp(*argv,"-q") == 0) p->verbose=0 ;
   else if (strcmp(*argv,"-v") == 0) p->verbose++ ;   
   else if (strcmp(*argv,"-r") == 0) {
     --argc ; ++argv ;
     if (!*argv) usage(prog,d) ;
    p->max_run = argint(prog,argv[0],d) ;   
   } else if (strcmp(*argv,"-fr") == 0) {
     --argc ; ++argv ;
     if (!*argv) usage(prog,d) ;
    p->max_run *= argdouble(prog,argv[0],d) ;   
   } else if (strcmp(*argv,"-s") == 0) {
     --argc ; ++argv ;
     if (!*argv) usage(prog,d) ;
     p->size_of_test = argint(prog,argv[0],d) ;   
   } else if (strcmp(*argv,"-fs") == 0) {
     --argc ; ++argv ;
     if (!*argv) usage(prog,d) ;
     p->size_of_test *= argdouble(prog,argv[0],d) ;   
   } else if (strcmp(*argv,"-f") == 0) {
     --argc ; ++argv ;
     if (!*argv) usage(prog,d) ;
     double f = argdouble(prog,argv[0],d) ;
     p->size_of_test *= f ;     
     p->max_run /= f ;
   } else if (strcmp(*argv,"-n") == 0) {
     --argc ; ++argv ;
     if (!*argv) usage(prog,d) ;
     p->n_exe = argint(prog,argv[0],d) ;
     if (p->n_exe < 1) p->n_exe = 1 ;
   } else if (strcmp(*argv,"-a") == 0) {
     --argc ; ++argv ;
     if (!*argv) usage(prog,d) ;
     int a = argint(prog,argv[0],d) ;
     p->avail = a ;
   } else if (strcmp(*argv,"-k") == 0) {
     --argc ; ++argv ;
     if (!*argv) usage(prog,d) ;
     int a = argint(prog,argv[0],d) ;
     p->sync_n = a < 0 ? 0 : a ;
   } else if (d->aff_incr >= 0 && strcmp(*argv,"-i") == 0) {
     --argc ; ++argv ;
     if (!*argv) usage(prog,d) ;
     int i = argint(prog,argv[0],d) ;
     p->aff_incr = i < 0 ? 0 : i ;
   } else if (d->aff_incr && strcmp(*argv,"-p") == 0) {
     --argc ; ++argv ;
     if (!*argv) usage(prog,d) ;
     cpus_t *cpus = argcpus(prog,argv[0],d) ;
     p->aff_cpus = cpus ;
   } else usage(prog,d) ;
  }

  if (p->max_run < 1) p->max_run = 1 ;
  if (p->size_of_test < 1) p->size_of_test = 1 ;

  /* Argument */
  if (argc == 0) return ;
  usage(prog,d) ;
}

/*************************/
/* Concurrency utilities */
/*************************/

/* phread based mutex */

pm_t *pm_create(void) {
  pm_t *p = malloc_check(sizeof(*p)) ;
  if (pthread_mutex_init(p,NULL)) { err(1,"mutex_init") ; }
  return p ;
}

void pm_free(pm_t *p) {
  free(p) ;
}

void pm_lock(pm_t *m) {
  if (pthread_mutex_lock(m)) { err(1,"mutex_lock") ; }
}

void pm_unlock(pm_t *m) {
  if (pthread_mutex_unlock(m)) { err(1,"mutex_unlock") ; }
}

/* phread condition */

pc_t *pc_create(void) {
  pc_t *p = malloc_check(sizeof(*p)) ;
  p->c_mutex = pm_create() ;
  p->c_cond = malloc_check(sizeof(*(p->c_cond))) ;
  if (pthread_cond_init(p->c_cond,NULL)) {  err(1,"cond_init") ; }
  return p ;
}

void pc_free(pc_t *p) {
  pm_free(p->c_mutex) ;
  free(p->c_cond) ;
  free(p) ;
}

static void pc_lock(pc_t *p) {
  pm_lock(p->c_mutex) ;
}

static void pc_unlock(pc_t *p) {
  pm_unlock(p->c_mutex) ;
}

void pc_wait(pc_t *p) {
  if (pthread_cond_wait(p->c_cond, p->c_mutex)) err(1,"cond_wait") ;
}

void pc_broadcast (pc_t *p) {
  if (pthread_cond_broadcast(p->c_cond)) err(1,"cond_broadcast") ;
}

static void pc_signal(pc_t *p) {
  if (pthread_cond_signal(p->c_cond)) err(1,"cond_signal") ;
}
/* pthread based barrier, usable for nproc threads */


pb_t *pb_create(int nprocs) {
  pb_t *p = malloc_check(sizeof(*p)) ;
  p->cond = pc_create() ;
  p->count = p->nprocs = nprocs ;
  return p ;
}

void pb_free(pb_t *p) {
  pc_free(p->cond) ;
  free(p) ;
}

void pb_wait(pb_t *p) {
  pc_lock(p->cond) ;
  --p->count ;
  if (p->count == 0) {
    p->count = p->nprocs ;
    pc_broadcast(p->cond) ;
  } else {
    pc_wait(p->cond) ;
  }
  pc_unlock(p->cond) ;
}

/* pthread based or flag */

po_t *po_create(int nprocs) {
  po_t *p = malloc_check(sizeof(*p)) ;
  p->cond = pc_create() ;
  p->nprocs = p->count = nprocs ;
  p->val = 0 ;
  return p ;
}

void po_free(po_t *p) {
  pc_free(p->cond) ;
  free(p) ;
}

void po_reinit(po_t *p) {
  pc_lock(p->cond) ;
  --p->count ;
  if (p->count == 0) {
    p->count = p->nprocs ;
    p->val = 0 ;
    pc_broadcast(p->cond) ;
  } else {
    pc_wait(p->cond) ;
  }
  pc_unlock(p->cond) ;
}

int po_wait(po_t *p, int v) {
  pc_lock(p->cond) ;
  --p->count ;
  p->val = p->val || v ;
  if (p->count == 0) {
    p->count = p->nprocs ;
    pc_broadcast(p->cond) ;
  } else {
    pc_wait(p->cond) ;
  }
  int r = p->val ;
  pc_unlock(p->cond) ;
  return r ;
}


/* One place buffer */

op_t *op_create(void) {
  op_t *p = malloc_check(sizeof(*p)) ;
  p->cond = pc_create() ;
  p->val = NULL ;
  p->some = 0 ;
  return p;
}

void op_free(op_t *p) {
  pc_free(p->cond) ;
  free(p) ;
}

void op_set(op_t *p, void *v) {
  pc_lock(p->cond) ;
  if (p->some) err(1,"op_set") ;
  p->val = v ;
  p->some = 1 ;
  pc_signal(p->cond) ;
  pc_unlock(p->cond) ;
}

void *op_get(op_t *p) {
  void *v = NULL ;
  pc_lock(p->cond) ;
  while (!p->some) {
    pc_wait(p->cond) ;
  }
  v = (void *) p->val ;
  p->val = NULL ;
  p->some = 0 ;
  pc_unlock(p->cond) ;
  return v ;
}

/* Thread launch and join */

void launch(pthread_t *th, f_t *f, void *a) {
  if (pthread_create(th,NULL,f,a)) { err(1,"phread_create"); }
}

void *join(pthread_t *th) {
  void *r ;
  if (pthread_join (*th,&r)) {err(1,"pthread_join"); }
  return r ;
}

/* Thread cache */

void *start_thread(void *_a) {
  sarg_t *_b = (sarg_t *)_a ;
  for (int _k = _b->max_run ; _k > 0 ; _k--) {
    void *_c = op_get(_b->op_arg) ;
    f_t *f = (f_t *)_c ;
    if (f == NULL) break ;
    void *ret = f(_b->arg) ;
    op_set(_b->op_ret,ret) ;
  }
  return NULL ;
}

/*****************/
/* Random things */
/*****************/


/*
  Having a fast (unlocked) rand function is critical for speed.
  Quality of generator is less important.
*/

/*
  Simple generator
  http://en.wikipedia.org/wiki/Linear_congruential_generator
*/


/*

From ocaml sources: (globroot.c)
Linear congruence with modulus = 2^32, multiplier = 69069
(Knuth vol 2 p. 106, line 15 of table 1), additive = 25173.


Knuth (vol 2 p. 13) shows that the least significant bits are
"less random" than the most significant bits with a modulus of 2^m.
We just swap half words, enough? */

static const unsigned a = 69069;
static const unsigned c = 25173 ;
// static unsigned st = 0;

inline static unsigned unlocked_rand(st_t *st)  {
  unsigned r = a * *st + c ;
  *st = r ;
  /* Swap high & low bits */
  unsigned int low = r & 0xffff ;
  unsigned int high = r >> 16 ;
  r = high | (low << 16) ;
  return r ;
}

int rand_bit(st_t *st)  {
  unsigned r = unlocked_rand(st) ;
  r &= 1 ;
  return r ; 
}

static const unsigned r_max = UINT_MAX ;

inline static unsigned rand_k (unsigned *st,unsigned k) {
  unsigned r, v ;
  do {
    r = unlocked_rand(st) ;
    v = r % k ;
  } while (r-v > r_max-k+1) ;
  return v ;
}

void perm_ints(unsigned *st,int *_t, int n) {
  int k;
  for (k = 0 ; k < n-1 ; k++) {
    int j = k+rand_k(st,n-k);
    int x = _t[k]; _t[k] = _t[j]; _t[j] = x;
  }
}

void perm_funs(unsigned *st,f_t *fun[], int n) {
  int k;
  for (k = 0 ; k < n-1 ; k++) {
    int j = k+rand_k(st,n-k);
    f_t *t = fun[j];
    fun[j] = fun[k]; fun[k] = t;
  }
}

void perm_ops(unsigned *st,op_t *op[], int n) {
  int k;
  for (k = 0 ; k < n-1 ; k++) {
    int j = k+rand_k(st,n-k);
    op_t *t = op[j];
    op[j] = op[k]; op[k] = t;
  }
}

void perm_threads(unsigned *st,pthread_t thread[], int n) {
  int k;
  for (k = 0 ; k < n-1 ; k++) {
    int j = k+rand_k(st,n-k);
    pthread_t t = thread[j];
    thread[j] = thread[k]; thread[k] = t;
  }
}

/****************/
/* Time counter */
/****************/

#include <sys/time.h>
#include <time.h>

tsc_t timeofday(void) {
  struct timeval tv ;
  if (gettimeofday(&tv,NULL)) err(1,"gettimeoday") ;  
  return tv.tv_sec * ((tsc_t)1000000) + tv.tv_usec ;
}

double tsc_ratio(tsc_t t1, tsc_t t2) {
  return ((double) t1) / ((double)t2) ;
}


double tsc_millions(tsc_t t) {
  return t / 1000000.0 ;
}
