(*********************************************************************)
(*                          Litmus                                   *)
(*                                                                   *)
(*        Luc Maranget, INRIA Paris-Rocquencourt, France.            *)
(*        Susmit Sarkar, University of Cambridge, UK.                *)
(*                                                                   *)
(*  Copyright 2010 Institut National de Recherche en Informatique et *)
(*  en Automatique and the authors. All rights reserved.             *)
(*  This file is distributed  under the terms of the Lesser GNU      *)
(*  General Public License.                                          *)
(*********************************************************************)

open Globals
module type I = sig
  type arch_reg
  val arch : Misc.arch
  val forbidden_regs : arch_reg list
  val reg_compare : arch_reg -> arch_reg -> int
  val reg_to_string : arch_reg -> string
(* Initial value of internal registers *)
  val internal_init : arch_reg -> string option

end


exception Error of string

module type S = sig
  type arch_reg 
  type v 

  type flow = Next | Branch of string
  type ins =
      { memo:string ; inputs:arch_reg list ;  outputs:arch_reg list;
        (* Jumps *)
        label:string option ;  branch : flow list ; }

  val empty_ins : ins 

  type t = {
      init : (arch_reg * v) list ;
      addrs : (int * string) list ;
      final : arch_reg list ;
      code : ins list;
    }
  val get_addrs : t -> string list
  val fmt_reg : arch_reg -> string
  val dump_label : string -> string
  val dump_out_reg : int -> arch_reg -> string
  val dump_v : v -> string
  val dump_casted : v -> string
  val dump : out_channel -> int -> t  -> unit
end

module Make(A:I) (V:Constant.S): S
  with type arch_reg = A.arch_reg
  and type v = V.v =
struct
  open Printf

  type arch_reg = A.arch_reg
  type v = V.v

  type flow = Next | Branch of string
  type ins =
      { memo:string ; inputs:arch_reg list ;  outputs:arch_reg list;
        (* Jumps *)
        label:string option ;  branch : flow list ; }

  let empty_ins =
    { memo="" ; inputs=[]; outputs=[];
      label=None; branch=[Next]; }

  type t = {
      init : (arch_reg * v) list ;
      addrs : (int * string) list ;
      final : arch_reg list ;
      code : ins list;
    }


  let get_addrs { init=init; addrs=addrs; } =
    let set =
      StringSet.union
        (StringSet.of_list (List.map (fun (_,a) -> a) addrs))
        (StringSet.of_list
           (List.fold_left
              (fun k (_,v) ->
                match v with V.Symbolic s -> s::k
                | V.Concrete _ -> k)
              [] init)) in
    StringSet.elements set
  
  
  exception Internal of string
  let internal msg = raise (Internal msg)

  exception Error of string
  let error msg = raise (Error msg)


  module RegSet =
    MySet.Make
      (struct
        type t = A.arch_reg
        let compare = A.reg_compare
      end)


  let escape_percent s =
    Misc.map_string
      (fun c -> match c with
      | '%' -> "%%" 
      | _ -> String.make 1 c)
      s

  let pp_reg r = escape_percent (A.reg_to_string r)
  let fmt_reg = pp_reg

  let dump_label lbl = "Lit" ^ lbl

  let clean_reg s =
    Misc.map_string
      (fun c -> match c with
      | '%' -> ""
      | _  -> String.make 1 c)
      s

  let tag_reg reg = clean_reg (A.reg_to_string reg)

  let tag_reg_ref reg = sprintf "%%[%s]" (tag_reg reg)
  and tag_reg_def reg = sprintf "[%s]" (tag_reg reg)

  let dump_out_reg proc reg =
    sprintf "out_%i_%s"
      proc
      (clean_reg (A.reg_to_string reg))

  let dump_trashed_reg reg =
    sprintf "trashed_%s"
      (clean_reg (A.reg_to_string reg))

  let compile_out_reg proc reg =
    sprintf "%s[_i]" (dump_out_reg proc reg)


  let get_reg k rs =
    try List.nth rs k
    with _ ->
      internal
        (sprintf "get_reg %i in {%s}"
           k (String.concat ","
                (List.map pp_reg rs)))


  let to_string t =

    let digit i =
      let c = Char.code t.memo.[i] in
      let n = c - Char.code '0' in
      if 0 <= n && n <= 2 then n
      else internal (sprintf "bad digit '%i'" n)
    
    and substring i j = 
      try String.sub t.memo i (j-i)
      with _ -> internal (sprintf "substring %i-%i" i j)

    and look_escape i =
      try String.index_from t.memo i '^'
      with
      | Not_found -> raise Not_found
      | _ -> internal (sprintf "look_escape %i" i) in
      

    let b = Buffer.create 20 in
    let add = Buffer.add_string b in
    let len = String.length t.memo in

    let rec do_rec i = 
      if i < len then
        try
          let j = look_escape i in
          add (substring i j) ;
          let n = digit (j+2) in
          begin match t.memo.[j+1] with
          | 'i' -> add (tag_reg_ref (get_reg n t.inputs))
          | 'o' -> add (tag_reg_ref (get_reg n t.outputs))
          | c -> internal (sprintf "bad escape '%c'" c)
          end ;
          do_rec (j+3)
        with Not_found -> add (substring i len) in
    try do_rec 0  ; Buffer.contents b
    with Internal msg ->
      error (sprintf "memo: %s, error: %s" t.memo msg)

  let dump_outputs chan proc t trashed =
    let outs =
      String.concat ","
        (List.map
           (match get_memory A.arch with
           | Direct ->
               (fun (_,a) -> sprintf "[%s] \"=m\" (%s[_i])" a a)
           | Indirect ->
               (fun (_,a) -> sprintf "[%s] \"=m\" (*%s[_i])" a a)
           )
           t.addrs
         @List.map
             (fun reg -> sprintf "%s \"=&r\" (%s)"
                 (tag_reg_def reg) (compile_out_reg proc reg))
             t.final
         @RegSet.fold
             (fun reg k ->
               sprintf "%s \"=&r\" (%s)"
                 (tag_reg_def reg) (dump_trashed_reg reg)::k)
             trashed []) in
    fprintf chan ":%s\n" outs


  let dump_addr a = match get_memory A.arch with
  | Direct -> sprintf "&%s[_i]" a
  | Indirect -> sprintf "%s[_i]" a

  let dump_v v = match v with
  | V.Concrete i -> sprintf "%i" i
  | V.Symbolic a -> dump_addr a

(*
  let typeof v = match v with
  | V.Concrete _ -> "int"
  | V.Symbolic _ -> "int *"
*)
  let dump_casted v = match v with
  | V.Concrete i -> sprintf "%i" i
  | V.Symbolic a -> sprintf "((int *)%s)" (dump_addr a)

  let all_regs t =
    let all_ins ins =
      RegSet.union (RegSet.of_list (ins.inputs@ins.outputs)) in
    List.fold_right all_ins t.code  (RegSet.of_list t.final)

  let trashed_regs t =
    let trashed_ins ins = RegSet.union (RegSet.of_list ins.outputs) in
    let all_trashed =
      List.fold_right trashed_ins t.code RegSet.empty in
    RegSet.diff all_trashed (RegSet.of_list t.final)

  let pp_reg chan r = fprintf chan "%s" (tag_reg r)

  let dump_inputs chan t trashed =    
    let all = all_regs t in
    let in_outputs = RegSet.union trashed  (RegSet.of_list t.final) in      
(*
    eprintf "Outputs in In: %a\n"
      (fun chan rs -> RegSet.pp chan "," pp_reg rs)
      in_outputs ;
*)
    let dump_pair reg v =
      if RegSet.mem reg in_outputs then begin
        let r = sprintf "\"%s\" (%s)" (tag_reg_def reg) (dump_v v) in
(*        eprintf "Here: %s\n" r ; *)
        r
      end else match A.internal_init reg with
      | None ->
          sprintf "%s \"r\" (%s)" (tag_reg_def reg) (dump_v v)
      | Some s ->
          sprintf "%s \"r\" (%s)" (tag_reg_def reg) s in
          
    (* Input from state *)
    let ins =
      List.map
        (fun (reg,v) -> dump_pair reg v)
        t.init in
    (* All other inputs, apparently needed to get gcc to
       allocate registers avoiding all registers in template *)
    let rem =
      RegSet.diff
        all
        (List.fold_right
           (fun (reg,_) -> RegSet.add reg) t.init in_outputs) in
    let rem =
      RegSet.fold
        (fun reg k ->
          dump_pair reg (V.Concrete 0)::k)
        rem [] in

    fprintf chan ":%s\n" (String.concat "," (ins@rem))


  let dump_clobbers chan _t =
    fprintf chan ":%s\n"
      (String.concat ","
         (List.map (fun s -> sprintf "\"%s\"" s)
            ("cc"::"memory"::
             List.map A.reg_to_string A.forbidden_regs)))

  let dump chan proc t =
    let rec dump_ins k ts = match ts with
    | [] -> ()
    | t::ts ->
        fprintf chan "\"_litmus_P%i_%i_: %s\\n\\t\"\n"
          proc k (to_string t) ;
        dump_ins (k+1) ts in
    let trashed = trashed_regs t in
    RegSet.iter
      (fun reg ->
        fprintf chan "        int %s;\n"
          (dump_trashed_reg reg))
      trashed ;          
    fprintf chan "asm volatile (\n" ;
    begin match t.code with
    | [] -> fprintf chan "\"\"\n"
    | code -> dump_ins 0 code
    end ;
    dump_outputs chan proc t trashed ;
    dump_inputs chan t trashed ;
    dump_clobbers chan t  ;
    fprintf chan ");\n"
      
end
