(*********************************************************************)
(*                          Litmus                                   *)
(*                                                                   *)
(*        Luc Maranget, INRIA Paris-Rocquencourt, France.            *)
(*        Susmit Sarkar, University of Cambridge, UK.                *)
(*                                                                   *)
(*  Copyright 2010 Institut National de Recherche en Informatique et *)
(*  en Automatique and the authors. All rights reserved.             *)
(*  This file is distributed  under the terms of the Lesser GNU      *)
(*  General Public License.                                          *)
(*********************************************************************)

(***********************************************)
(* Parse a source file for the needs of litmus *)
(* (And then compile test)                     *)
(***********************************************)

open Printf

type answer =
  | Completed of Misc.arch * string * string list * StringSet.t
  | Interrupted of Misc.arch * exn
  | Absent of Misc.arch

module Make (A:Arch.S)
    (L:GenParser.LexParse with type instruction = A.pseudo)
    (XXXComp : XXXCompile.S with module A = A) =
  struct
    module P = GenParser.Make(A) (L)
    module T = Test.Make(A)
    module Comp = Compile.Make (T) (XXXComp)
    module S = Skel.Make(T)
    module R = Run.Make(T)
    open GenParser

    let get_cycle t =
      let info = t.MiscParser.info in
      List.assoc "Cycle" info

    let cycle_ok avoid t =
      try
        let cy = get_cycle t in
        not (avoid cy)
      with Not_found -> true

    let is_cycle_new cycles t =
      try
        let cy = get_cycle t in
        if StringSet.mem cy cycles then
          false,cycles
        else
          true,StringSet.add cy cycles
      with Not_found -> true,cycles

    let compile avoid_cycle utils cycles name in_chan out_chan splitted  =
      try begin
        let parsed = P.parse in_chan splitted
        and doc = splitted.Splitter.name in
        close_in in_chan ;
        let nprocs = List.length parsed.MiscParser.prog
        and avail = match Globals.get_avail A.arch with
        | Some n -> n
        | None -> 1000 in
        let is_new,cycles = is_cycle_new cycles parsed in
        if
          is_new &&
          cycle_ok avoid_cycle parsed &&
          (not (Globals.get_limit ()) || nprocs <= avail)
        then begin
          let module Alloc = SymbReg.Make(A) in
          let allocated = Alloc.allocate_regs parsed in
          let compiled = Comp.compile allocated in
          let source =
            let src = MyName.outname name ".c" in
            if Globals.is_tar () then Tar.outname src
            else src in
          let () = Misc.output_protect  (S.dump doc compiled) source in
          let utils =
            match utils with
            | [] ->
                let module O = ObjUtil.Make(A) in
                O.dump ()
            | _ -> utils  in
          R.run name out_chan doc allocated source ;
          Completed (A.arch,source,utils,cycles)
        end else begin
          Warn.warn
            "Test %s not performed" doc.Name.name ;
          Absent A.arch
        end
      end with e -> Interrupted (A.arch,e)

  end


let from_chan avoid_cycle fst cycles name in_chan out_chan =  
(* First split the input file in sections *)
  let { Splitter.arch=arch } as splitted =  Splitter.split name in_chan in
(* Then call appropriate compiler, depending upon arch *)
  let module V = SymbConstant in
    match arch with
    | Misc.PPC ->
        let module Conf = struct
          let word = Globals.get_word Misc.PPC
          let sync_macro =Globals.get_sync_macro ()
        end in
	let module PPC = PPCArch.Make(V) in
	let module PPCLexParse = struct
	  type instruction = PPC.pseudo
	  type token = PPCParser.token

	  let lexer = PPCLexer.token
	  let parser = PPCParser.main
	end in
        let module X = Make
            (PPC) (PPCLexParse) (PPCCompile.Make(V)(Conf)) in
        X.compile avoid_cycle fst cycles name in_chan out_chan splitted
    | Misc.X86 ->
	let module X86 = X86Arch.Make(V) in
	let module X86LexParse = struct
	  type instruction = X86.pseudo
	  type token = X86Parser.token

	  let lexer = X86Lexer.token
	  let parser = X86Parser.main
	end in
        let module X = Make
            (X86) (X86LexParse) (X86Compile.Make(V)) in
        X.compile avoid_cycle fst cycles name in_chan out_chan splitted
    | Misc.ARM ->  Warn.fatal "ARM not implemented"


let from_file avoid_cycles fst cycles name out_chan =
  Misc.input_protect
    (fun in_chan ->
      from_chan avoid_cycles fst cycles name in_chan out_chan) name

open MySys

let report_failure name msg chan =
  let title = sprintf "!!! Test %s failed !!!" name in
  let warn = String.make (String.length title) '!' in
  output_line chan  warn ;
  output_line chan  title ;
  output_line chan  warn ;
  fprintf chan "Reported: %s\n" msg ;
  output_line chan ""

let read_no fname =
  Misc.input_protect
    (fun chan -> MySys.read_list chan (fun s -> Some s))
    fname

let from_files names =  
  let avoid_cycles =
    let xs = match Globals.get_no () with
    | None -> []
    | Some fname -> read_no fname in
    let set = StringSet.of_list xs in
    fun cy -> StringSet.mem cy set in
  let out_chan =
    if Globals.is_tar() then
      open_out (Tar.outname (MyName.outname "run" ".sh"))
    else stdout in
  let exp = match Globals.get_exp () with
  | None -> None
  | Some exp -> Some (open_out exp) in
  if Globals.is_tar () then begin
    output_line out_chan "date" ;
    output_line out_chan "LITMUSOPTS=\"${@:-$LITMUSOPTS}\"" ;
    begin match Globals.get_para () with
    | Globals.Shell ->
        output_line out_chan "TMPDIR=${TMPDIR:-/var/tmp}" ;
        output_line out_chan "TMPDIR=$TMPDIR/tmp.$$" ;
        output_line out_chan "mkdir $TMPDIR"
    | Globals.Self -> ()
    end
  end ;
  let a,sources,utils,_ =
    List.fold_left
      (fun (_,srcs,fst,cycles) name ->
        match from_file avoid_cycles fst cycles name out_chan with
        | Completed (a,src,fst,cycles) ->
            begin match exp with
            | None -> ()
            | Some exp -> fprintf exp "%s\n" name
            end ;
            a,(src::srcs),fst,cycles
        | Absent a -> a,srcs,fst,cycles
        | Interrupted (a,e) ->
            let msg =  match e with
            | Misc.Exit -> "None"
            | Misc.Fatal msg 
            | Misc.UserError msg ->
                Pos.pp_pos0 stderr name ;
                Printf.eprintf "%s\n%!" msg ;
                msg
            | e -> raise e  in
            report_failure name msg out_chan ;
            a,srcs,fst,cycles)
      (Misc.X86,[],[],StringSet.empty)  names in

  begin match exp with
  | None -> ()
  | Some exp -> close_out exp
  end ;

  if (Globals.is_tar ()) then begin
    match Globals.get_para () with
    | Globals.Shell -> output_line out_chan "rmdir $TMPDIR"
    | Globals.Self -> ()
  end ;
  Run.report_machine a out_chan ;
  if (Globals.is_tar ()) then begin
    output_line out_chan "head -1 comp.sh" ;
    output_line out_chan "echo \"LITMUSOPTS=$LITMUSOPTS\"" ;
    output_line out_chan "date"
  end ;
  if (not (Globals.is_tar ())) then begin
    List.iter MySys.remove utils        
  end ;
  flush out_chan ;


  if Globals.is_tar () then begin
    close_out out_chan ;
    let sources = List.map Filename.basename  sources in
(* Shell script for sequential compilation *)
    Misc.output_protect
      (fun chan ->
        let gcc_opts = Run.get_gcc_opts a in
        fprintf chan "GCCOPTS=\"%s\"\n" gcc_opts ;
        let link_opts = Run.get_link_opts a in
        fprintf chan "LINKOPTS=\"%s\"\n" link_opts ;
        fprintf chan "/bin/rm -f *.exe *.s\n" ;
        List.iter
          (fun s ->
            if Filename.check_suffix s ".c" then
              fprintf chan "gcc $GCCOPTS -O2 -c %s\n"
                (Filename.basename s))
          utils ;
        let utils_objs =
          List.fold_right
            (fun s k ->
              if Filename.check_suffix s ".c" then
                let b = Filename.chop_suffix (Filename.basename s) ".c" in
                (b ^ ".o") :: k
              else k) utils [] in
        let utils_objs = String.concat " " utils_objs in
        List.iter
          (fun src ->
            let exe = Filename.chop_extension src ^ ".exe" in
(* No more moderate parallelism [blocks on abducens, sh bug ?] *)
            fprintf chan
              "gcc $GCCOPTS $LINKOPTS -o %s %s %s\n" exe utils_objs src ;
            let srcS = Filename.chop_extension src ^ ".s" in
            fprintf chan "gcc $GCCOPTS -S %s && mv %s %s.tmp && grep _litmus_ %s.tmp > %s && /bin/rm %s.tmp\n"
              src srcS srcS srcS srcS srcS ;
            ())
          (List.rev sources))
      (Tar.outname (MyName.outname "comp" ".sh")) ;
(* Makefile for parallel compilation *)
    if not (Globals.cross_mode ()) then 
      Misc.output_protect
        (fun chan ->
(* Variables *)
          let gcc_opts = Run.get_gcc_opts a in
          fprintf chan "GCCOPTS=%s\n" gcc_opts ;
          let link_opts = Run.get_link_opts a in
          fprintf chan "LINKOPTS=%s\n" link_opts ;
          fprintf chan "SRC=\\\n" ;
          List.iter
            (fun src -> fprintf chan " %s\\\n" src)
            (List.rev sources) ;
          fprintf chan "\n" ;
          fprintf chan "EXE=$(SRC:.c=.exe)\n" ;
          fprintf chan "S=$(SRC:.c=.s)\n" ;
          fprintf chan "\n" ;
(* Entry points *)
          fprintf chan "all: $(EXE) $(S)\n" ;
          fprintf chan "\n" ;
          fprintf chan "clean:\n" ;
          fprintf chan "\t/bin/rm -f *.o *.s *.exe *~\n" ;
          fprintf chan "\n" ;
(* Utilities *)
          let utils =
            List.fold_right
              (fun s k ->
                if Filename.check_suffix s ".c" then
                  let b = Filename.chop_suffix (Filename.basename s) ".c" in
                  b :: k
                else k) utils [] in          
          List.iter
            (fun u ->
              let src = u ^ ".c" and obj = u ^ ".o" in
              fprintf chan "%s: %s\n" obj src ;
              fprintf chan "\tgcc $(GCCOPTS) -O2 -c %s\n" src ;
              fprintf chan "\n")
            utils ;
(* EXE pattern rule *)
          let utils_objs =
            String.concat " " (List.map (fun s -> s ^ ".o") utils) in
          fprintf chan "%%.exe:%%.c %s\n" utils_objs ;
          fprintf chan
            "\tgcc $(GCCOPTS) $(LINKOPTS) -o $@ %s $<\n" utils_objs ;
          fprintf chan "\n" ;
(* S pattern rule *)
          fprintf chan "%%.s:%%.c\n" ;
          fprintf chan "\tgcc $(GCCOPTS) -S $< && mv $@ $@.tmp && \\\n" ;
          fprintf chan "\tgrep _litmus_ $@.tmp > $@ && /bin/rm $@.tmp\n" ;
          fprintf chan "\n")
        (Tar.outname (MyName.outname "Makefile" "")) ;
    if Globals.cross_mode () then begin
(*      ObjUtil.cp_lib_file 
        "Makefile-for-cross" (Tar.outname (MyName.outname "Makefile" "")) ; *)
      Misc.output_protect 
        (fun chan -> output_line chan (GenerateCrossDoc.gen_makefile a))
        (Tar.outname (MyName.outname "Makefile" "")) ; 
      ObjUtil.cp_lib_file
        "README-for-cross" (Tar.outname (MyName.outname "README" "")) ;
      let tar = Globals.get_tar () in
      let sources = List.map Filename.basename sources in
      let utils = List.map Filename.basename utils in
      Tar.tar_structured_dir tar "litmus_tests"
        [("src",["comp.sh";"run.sh"] @ sources @ utils)]
    end else
      Tar.tar  (Globals.get_tar ())
  end ;
  ()
    

